.md-typeset .admonition .md-typeset__table table > thead {
    display: none;
}

.md-typeset .admonition .md-typeset__table table > tbody > tr > td:first-child {
    white-space: nowrap;
    max-width: 1%;
    padding-left: 1.5em;
    padding-right: 4px;
}

.md-typeset .admonition .md-typeset__table table {
    box-shadow: inherit;
}

.md-typeset .admonition .md-typeset__table table th {
    background-color: hsla(198, 100%, 95%, .40);
    color: inherit;
    font-weight: bold;
}

.md-typeset .admonition .md-typeset__table table > tbody > tr > td:nth-child(2) {
    white-space: nowrap;
    max-width: 1%;
    padding-left: 4px;
    padding-right: 4px;
}
.md-typeset .admonition .md-typeset__table table > tbody > tr > td:nth-child(3) {
    width: 100%;
}

.md-typeset .admonition .md-typeset__table {
    width: 100%;
    border-radius: 0;
    padding: 0;
    margin-top: 0;
    margin-bottom: 0;
}

.md-typeset .admonition .md-typeset__scrollwrap {
    margin-top: 0;
    margin-bottom: 0;
    padding-top: 0;
    padding-bottom: 0;
}

html .md-typeset .admonition > :last-child:not(p) {
    margin-bottom: 0;
}

.admonition pre {
    margin: 0;
}

.md-typeset .admonition {
    padding-left: 0;
    padding-right: 0;
}

.md-typeset .admonition-title {
    margin-left: 0;
}

.md-typeset .admonition p:not([class]) {
    margin-left: 0;
    padding-left: 8px;
}
